from abc import ABC, abstractmethod
from datetime import datetime, date
from typing import Any, Iterable, Tuple


class PagesSorter(ABC):
    _default_published = date.fromtimestamp(0)

    def __init__(self, pages: Iterable[dict]):
        self.pages = pages

    @abstractmethod
    def __call__(self, page: dict) -> Any:
        raise NotImplemented()


class PagesSortByTime(PagesSorter):
    def __call__(self, page: dict) -> datetime:
        return page.get('published', self._default_published)


class PagesSortByFolderAndTime(PagesSorter):
    def __call__(self, page: dict) -> Tuple:
        return (
            page.get('folder'),
            date.today() - page.get(
                'published', self._default_published
            )
        )


class PagesSortByTimeGroupedByFolder(PagesSorter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        st = {}
        for page in self.pages:
            folder = page.get('folder', '')
            published = page.get('published', self._default_published)
            st[folder] = st.get(folder, published)
            st[folder] = max(st[folder], published)

        self._max_date_by_folder = st

    def __call__(self, page: dict) -> Tuple:
        return (
            self._max_date_by_folder[page.get('folder', '')],
            page.get('published', self._default_published)
        )



import os
import sys
import logging
import fcntl

from .cli import get_args
from .config import init_config

# When running under gunicorn/uWSGI, sys.argv usually belongs to the process
# manager, not to madblog. Prefer environment variables for configuration.
env_blog_dir = os.environ.get("MADBLOG_CONTENT_DIR")
env_config_file = os.environ.get("MADBLOG_CONFIG")

opts = None
blog_dir = env_blog_dir
config_file = env_config_file

if not blog_dir and not config_file:
    try:
        parsed, _ = get_args(sys.argv[1:])
        # Gunicorn/uWSGI may pass arguments like `-w 8` - `8` would otherwise be
        # interpreted as the positional blog dir and override config.content_dir.
        if parsed.dir and not os.path.isdir(parsed.dir):
            parsed.dir = None

        # Only trust the positional dir if it's a real directory.
        if parsed.dir and os.path.isdir(parsed.dir):
            opts = parsed
            blog_dir = parsed.dir

        # Only trust the config path if it points to an actual file.
        if parsed.config and os.path.isfile(os.path.expanduser(parsed.config)):
            opts = parsed
            config_file = os.path.expanduser(parsed.config)
    except Exception:
        pass

if not config_file:
    blog_dir = blog_dir or "."
    config_file = os.path.join(blog_dir, "config.yaml")

config = init_config(config_file=config_file, args=opts)

logging.basicConfig(
    level=logging.DEBUG if getattr(config, "debug", False) else logging.INFO,
    format="%(levelname)s: %(message)s",
)

from .app import app

# For gunicorn/uWSGI compatibility
application = app

_monitor_lock_f = None


def _start_monitor_once() -> None:
    if not getattr(config, "enable_webmentions", False):
        return

    global _monitor_lock_f

    lock_path = os.path.join(
        getattr(config, "content_dir", "."), ".madblog-webmentions-monitor.lock"
    )
    try:
        _monitor_lock_f = open(lock_path, "w")
        fcntl.flock(_monitor_lock_f, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except Exception:
        return

    try:
        app.start()
        logging.getLogger(__name__).info("Started webmentions filesystem monitor")
    except Exception:
        logging.getLogger(__name__).exception(
            "Failed to start webmentions filesystem monitor"
        )


try:
    import uwsgi  # type: ignore

    @uwsgi.postfork
    def _uwsgi_postfork():
        # Only start the monitor in one worker to avoid duplicated watchers.
        # (If you want one monitor per worker, remove this guard.)
        if getattr(uwsgi, "worker_id", lambda: 0)() != 1:
            return
        _start_monitor_once()

except Exception:
    # Not running under uWSGI - start when imported (e.g. gunicorn workers).
    _start_monitor_once()


# vim:sw=4:ts=4:et:



import logging
import smtplib
from dataclasses import dataclass
from datetime import datetime, timezone
from email.message import EmailMessage
from email.utils import format_datetime, make_msgid
from typing import Callable, Optional

from webmentions import Webmention, WebmentionDirection

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SmtpConfig:
    server: str
    port: int = 587
    username: Optional[str] = None
    password: Optional[str] = None
    starttls: bool = True
    enable_starttls_auto: bool = True
    sender: Optional[str] = None


def _send_email(
    *,
    smtp: SmtpConfig,
    recipient: str,
    subject: str,
    body: str,
) -> None:
    msg = EmailMessage()
    msg["To"] = recipient
    msg["From"] = smtp.sender or smtp.username or recipient
    msg["Subject"] = subject
    msg["Date"] = format_datetime(datetime.now(timezone.utc))
    msg["Message-ID"] = make_msgid(domain=None)
    msg.set_content(body)

    with smtplib.SMTP(smtp.server, smtp.port) as client:
        if smtp.enable_starttls_auto:
            client.ehlo()

        if smtp.starttls:
            client.starttls()
            if smtp.enable_starttls_auto:
                client.ehlo()

        if smtp.username:
            client.login(smtp.username, smtp.password or "")

        client.send_message(msg)


def build_webmention_email_notifier(
    *,
    recipient: str,
    blog_base_url: str,
    smtp: SmtpConfig,
    sender: Optional[str] = None,
    send_email: Callable[..., None] = _send_email,
) -> Callable[[Webmention], None]:
    smtp = SmtpConfig(
        server=smtp.server,
        port=smtp.port,
        username=smtp.username,
        password=smtp.password,
        starttls=smtp.starttls,
        enable_starttls_auto=smtp.enable_starttls_auto,
        sender=sender or smtp.sender,
    )

    def _on_mention_processed(mention: Webmention) -> None:
        if mention.direction != WebmentionDirection.IN:
            return

        subject = f"New Webmention received for {blog_base_url}"

        lines = [
            "A new Webmention has been processed.",
            "",
            f"Source: {mention.source}",
            f"Target: {mention.target}",
        ]

        if mention.author_name or mention.author_url:
            lines.append(
                f"Author: {mention.author_name or ''} {mention.author_url or ''}".strip()
            )

        if mention.title:
            lines.append(f"Title: {mention.title}")

        if mention.excerpt:
            lines.extend(["", "Excerpt:", mention.excerpt])

        try:
            send_email(
                smtp=smtp,
                recipient=recipient,
                subject=subject,
                body="\n".join(lines) + "\n",
            )
        except Exception:
            logger.exception(
                "Failed to send Webmention notification email (source=%s target=%s)",
                mention.source,
                mention.target,
            )

    return _on_mention_processed



import logging
import os
import re
from typing import Optional
from urllib.parse import urljoin

from feedgen.feed import FeedGenerator
from flask import (
    jsonify,
    request,
    Response,
    redirect,
    send_from_directory as send_from_directory_,
    render_template,
)

from .app import app
from .config import config
from ._sorters import PagesSortByTimeGroupedByFolder

logger = logging.getLogger(__name__)


def send_from_directory(
    path: str, file: str, alternative_path: Optional[str] = None, **kwargs
):
    if not os.path.exists(os.path.join(path, file)) and alternative_path:
        path = alternative_path
    return send_from_directory_(path, file, **kwargs)


@app.route("/", methods=["GET"])
def home_route():
    return render_template(
        "index.html",
        pages=app.get_pages(sorter=PagesSortByTimeGroupedByFolder),
        config=config,
    )


@app.route("/img/<img>", methods=["GET"])
def img_route(img: str):
    return send_from_directory(app.img_dir, img, config.default_img_dir)


@app.route("/favicon.ico", methods=["GET"])
def favicon_route():
    return img_route("favicon.ico")


@app.route("/js/<file>", methods=["GET"])
def js_route(file: str):
    return send_from_directory(app.js_dir, file, config.default_js_dir)


@app.route("/pwabuilder-sw.js", methods=["GET"])
def pwa_builder_route():
    return send_from_directory(app.js_dir, "pwabuilder-sw.js", config.default_js_dir)


@app.route("/pwabuilder-sw-register.js", methods=["GET"])
def pwa_builder_register_route():
    return send_from_directory(
        app.js_dir, "pwabuilder-sw-register.js", config.default_js_dir
    )


@app.route("/css/<style>", methods=["GET"])
def css_route(style: str):
    return send_from_directory(app.css_dir, style, config.default_css_dir)


@app.route("/fonts/<file>", methods=["GET"])
def fonts_route(file: str):
    return send_from_directory(app.fonts_dir, file, config.default_fonts_dir)


@app.route("/manifest.json", methods=["GET"])
def manifest_route():
    # If there is a manifest.json in the content directory, use it
    manifest_file = os.path.join(config.content_dir, "manifest.json")
    if os.path.isfile(manifest_file):
        return send_from_directory(config.content_dir, "manifest.json")

    # Otherwise, generate a default manifest.json
    return jsonify(
        {
            "name": config.title,
            "short_name": config.title,
            "icons": [
                {"src": "/img/icon-48.png", "sizes": "48x48", "type": "image/png"},
                {"src": "/img/icon-72.png", "sizes": "72x72", "type": "image/png"},
                {"src": "/img/icon-96.png", "sizes": "96x96", "type": "image/png"},
                {"src": "/img/icon-144.png", "sizes": "144x144", "type": "image/png"},
                {"src": "/img/icon-168.png", "sizes": "168x168", "type": "image/png"},
                {"src": "/img/icon-192.png", "sizes": "192x192", "type": "image/png"},
                {"src": "/img/icon-256.png", "sizes": "256x256", "type": "image/png"},
                {"src": "/img/icon-512.png", "sizes": "512x512", "type": "image/png"},
            ],
            "gcm_sender_id": "",
            "gcm_user_visible_only": True,
            "start_url": "/",
            "permissions": ["gcm"],
            "scope": "",
            "orientation": "portrait",
            "display": "standalone",
            "theme_color": "#000000",
            "background_color": "#ffffff",
        }
    )


@app.route("/article/<path:path>/<article>", methods=["GET"])
def article_with_path_route(path: str, article: str):
    return app.get_page(os.path.join(path, article))


@app.route("/article/<article>", methods=["GET"])
def article_route(article: str):
    return article_with_path_route("", article)


def _get_absolute_url(url: str) -> str:
    if not url:
        return ""

    if re.search(r"^https?://", url):
        return url

    return urljoin(config.link, url)


@app.route("/feed", methods=["GET"])
def feed_route():
    feed_type = request.args.get("type", "rss").lower().strip()
    if feed_type not in {"rss", "atom"}:
        return Response("Invalid feed type", status=400, mimetype="text/plain")

    short_description = "short" in request.args or config.short_feed
    pages = app.get_pages(
        with_content=not short_description,
        skip_header=True,
        skip_html_head=True,
    )

    pages = pages[: config.max_entries_per_feed]

    fg = FeedGenerator()
    fg.id(config.link)
    fg.title(config.title)
    fg.link(href=config.link, rel="alternate")
    fg.description(config.description)
    fg.language(config.language)

    for category in config.categories:
        fg.category(term=category)

    icon_url = _get_absolute_url("/img/icon.png")
    if icon_url:
        fg.logo(icon_url)

    self_url = _get_absolute_url(f"/feed?type={feed_type}")
    fg.link(href=self_url, rel="self")

    if pages and pages[0][1].get("published"):
        fg.updated(pages[0][1]["published"])

    for _, page in pages:
        uri = page.get("uri", "")
        entry_url = _get_absolute_url(uri)

        fe = fg.add_entry()
        if entry_url:
            fe.id(entry_url)
            fe.link(href=entry_url)

        fe.title(page.get("title", "[No Title]"))

        if page.get("published"):
            fe.published(page["published"])
            fe.updated(page["published"])

        if page.get("description"):
            fe.summary(page.get("description", ""))

        if not short_description:
            fe.content(page.get("content", ""), type="html")

        image_url = _get_absolute_url(page.get("image", ""))
        if image_url:
            fe.link(href=image_url, rel="enclosure")

    if feed_type == "atom":
        return Response(fg.atom_str(pretty=True), mimetype="application/atom+xml")

    return Response(fg.rss_str(pretty=True), mimetype="application/rss+xml")


@app.route("/rss", methods=["GET"])
def rss_route():
    """
    This route exists only for backward compatibility.

    It redirects to the /feed route with the appropriate query parameter to generate an RSS feed.
    """
    qs = request.query_string.decode("utf-8")
    if qs:
        return redirect(f"/feed?type=rss&{qs}")
    return redirect("/feed?type=rss")


# vim:sw=4:ts=4:et:
